/*
 * ivykis, an event handling library
 * Copyright (C) 2002, 2003, 2009 Lennert Buytenhek
 * Dedicated to Marija Kulikova.
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street - Fifth Floor, Boston,
 * MA 02110-1301, USA.
 */

#ifdef sun

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <syslog.h>
#include <sys/devpoll.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/poll.h>
#include "iv_private.h"

#define HASH_SIZE		(512)
#define UPLOAD_QUEUE_SIZE	(1024)

static struct pollfd		*batch;
static int			batch_size;
static struct list_head		*htable;
static int			poll_fd;
static struct pollfd		*upload_queue;
static int			upload_entries;


static unsigned int __fd_hash(unsigned int fd)
{
	return fd % HASH_SIZE;
}

static struct iv_fd_ *find_fd(int fd)
{
	int hash = __fd_hash(fd);
	struct list_head *lh;
	struct iv_fd_ *ret = NULL;

	list_for_each(lh, &htable[hash]) {
		struct iv_fd_ *f;

		f = list_entry(lh, struct iv_fd_, list_hash);
		if (f->fd == fd) {
			ret = f;
			break;
		}
	}

	return ret;
}


/* interface ****************************************************************/
static int iv_dev_poll_init(int maxfd)
{
	int i;

	poll_fd = open("/dev/poll", O_RDWR);
	if (poll_fd < 0)
		return -1;

	batch = malloc(maxfd * sizeof(*batch));
	if (batch == NULL) {
		close(poll_fd);
		return -1;
	}

	htable = malloc(HASH_SIZE * sizeof(*htable));
	if (htable == NULL) {
		free(batch);
		close(poll_fd);
		return -1;
	}

	upload_queue = malloc(UPLOAD_QUEUE_SIZE * sizeof(*upload_queue));
	if (upload_queue == NULL) {
		free(htable);
		free(batch);
		close(poll_fd);
		return -1;
	}

	for (i = 0; i < HASH_SIZE; i++)
		INIT_LIST_HEAD(&htable[i]);
	batch_size = maxfd;
	upload_entries = 0;

	return 0;
}

static void flush_upload_queue(void)
{
	unsigned char *base;
	int left;

	base = (unsigned char *)upload_queue;
	left = upload_entries * sizeof(struct pollfd);

	while (left) {
		int ret;

		do {
			ret = write(poll_fd, base, left);
		} while (ret < 0 && errno == EINTR);

		if (ret < 0) {
			syslog(LOG_CRIT, "flush_upload_queue: got error %d[%s]",
			       errno, strerror(errno));
			abort();
		}

		base += ret;
		left -= ret;
	}

	upload_entries = 0;
}

static void iv_dev_poll_poll(struct list_head *active, int msec)
{
	int i;
	int ret;

#if 0
	/*
	 * @@@ Is this necessary?
	 * @@@ This is ugly and dependent on clock tick granularity.
	 */
	if (msec)
		msec += (1000/100) - 1;
#endif

	if (upload_entries)
		flush_upload_queue();

	do {
		struct dvpoll dvp;

		dvp.dp_fds = batch;
		dvp.dp_nfds = batch_size;
		dvp.dp_timeout = msec;
		ret = ioctl(poll_fd, DP_POLL, &dvp);
	} while (ret < 0 && errno == EINTR);

	if (ret < 0) {
		syslog(LOG_CRIT, "iv_dev_poll_poll: got error %d[%s]",
		       errno, strerror(errno));
		abort();
	}

	for (i = 0; i < ret; i++) {
		struct iv_fd_ *fd;

		fd = find_fd(batch[i].fd);
		if (fd == NULL) {
			syslog(LOG_CRIT, "iv_dev_poll_poll: got event for "
					 "unknown fd %d", batch[i].fd);
			abort();
		}

		if (batch[i].revents & (POLLIN | POLLERR | POLLHUP))
			iv_fd_make_ready(active, fd, MASKIN);

		if (batch[i].revents & (POLLOUT | POLLERR))
			iv_fd_make_ready(active, fd, MASKOUT);

		if (batch[i].revents & POLLERR)
			iv_fd_make_ready(active, fd, MASKERR);
	}
}

static void iv_dev_poll_register_fd(struct iv_fd_ *fd)
{
	list_add_tail(&fd->list_hash, &htable[__fd_hash(fd->fd)]);
}

static void iv_dev_poll_unregister_fd(struct iv_fd_ *fd)
{
	list_del_init(&fd->list_hash);
}

static int bits_to_poll_mask(int bits)
{
	int mask;

	mask = 0;
	if (bits & MASKIN)
		mask |= POLLIN | POLLRDNORM | POLLRDBAND | POLLPRI;
	if (bits & MASKOUT)
		mask |= POLLOUT | POLLWRNORM | POLLWRBAND;
	if (bits)
		mask |= POLLERR | POLLHUP;

	return mask;
}

static void iv_dev_poll_notify_fd(struct iv_fd_ *fd, int wanted)
{
	if (upload_entries > UPLOAD_QUEUE_SIZE - 2)
		flush_upload_queue();

	if (fd->registered_bands & ~wanted) {
		upload_queue[upload_entries].fd = fd->fd;
		upload_queue[upload_entries].events = POLLREMOVE;
		upload_entries++;
	}

	if (wanted) {
		upload_queue[upload_entries].fd = fd->fd;
		upload_queue[upload_entries].events = bits_to_poll_mask(wanted);
		upload_entries++;
	}

	fd->registered_bands = wanted;
}

static void iv_dev_poll_deinit(void)
{
	free(upload_queue);
	free(htable);
	free(batch);
	close(poll_fd);
}


struct iv_poll_method iv_method_dev_poll = {
	.name		= "dev_poll",
	.init		= iv_dev_poll_init,
	.poll		= iv_dev_poll_poll,
	.register_fd	= iv_dev_poll_register_fd,
	.unregister_fd	= iv_dev_poll_unregister_fd,
	.notify_fd	= iv_dev_poll_notify_fd,
	.deinit		= iv_dev_poll_deinit,
};
#endif
